#include "DDHUndoableKnife.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>

// DOT product for non-pointer versions of vertex classes/structs
#define DDH_DOTPRODUCT(p1, p2)  ((float)((p1).x*(p2).x + (p1).y*(p2).y + (p1).z*(p2).z))
// Epsilon comparator
#define DDH_EPSILON 0.00001f
// If a slice appears too close to a vertex, we don't count it
#define DDH_SLICE_EPSILON 0.00005f

// Added by DDH to expose Slice Surface, change Select Mode
#ifdef __cplusplus
extern "C"
{
Prototype void selected_slice_surface_across_vertices( List *vertlist );
Prototype void set_new_select_mode( int mode );
Prototype void selected_surfaces_create_object();
}
#endif

//---------------------------------------------------------------------------------------
// Generic function for copying data from two lists to a third list
// Returns the number of items merged
int ddhlist_merge( List **out, List *l1, List *l2 )
{
   int ictr = 0;
   for( List *p1 = l1; p1 != NULL; p1 = p1->next )
   {
      for( List *p2 = l2; p2 != NULL; p2 = p2->next )
      {
         if( p1->data == p2->data )
         {
            list_add_item_head( out, p1->data );
            ++ictr;
            break;
         }
      }
   }
   return ictr;
}
//---------------------------------------------------------------------------------------
DDHUndoableKnife::DDHUndoableKnife( ACObject *ko, bool doslice, bool cutaway, bool delknife )
{

   mUndoCalled = false;

   // Knife elements
   mKnifeObject = ko;
   mSliceSurfacesInd = doslice;
   mCutAwayInd = cutaway;
   mDeleteKnifeInd = delknife;

   // Objects/data
   mDeletedKnife = NULL;
   mDeletedKnifeOwner = NULL;
   mPreKnife = NULL;
   mPostKnife = NULL;
   mPostSelectedObjects = NULL;

   // Knife currently works in object mode.
   // This is because we allow multiple cuts on an edge, and we need to
   // know where to select for our second pass.
   int selmode = ac_get_select_mode();
   if( selmode != SELECT_GROUP && selmode != SELECT_OBJECT )
   {
      message_dialog( "knife : must be in Object or Group mode" );
      abort();
      return;
   }
 
   // Ensure the knife object is valid
   if( !object_exists( mKnifeObject ))
   {
      message_dialog( "knife : knife is not specified or no longer exists" );
      abort();
      return;
   }
   // object_is_selected() doesn't link properly --- entry missing from ac3d.lib???
   // Have to iterate through each object to see if an object is selected
   List *so = ac_selection_get_objects();
   for( List *pso = so; pso != NULL; pso = pso->next )
   {
      if((( ACObject * )pso->data ) == mKnifeObject )
      {
         message_dialog( "knife : knife cannot slice itself - please deselect the knife" );
         abort();
         return;
      }
   }
   list_free( &so );
   // Abort if object has children
   if( ac_object_get_childrenlist( mKnifeObject ) != NULL )
   {
      message_dialog( "knife : knife must be a single object (with no child objects)" );
      abort();
      return;
   }
   // Abort if knife is not a poly
   if( ac_object_get_type( mKnifeObject ) != OBJECT_NORMAL )
   {
      message_dialog( "knife : knife must be a polygonal object" );
      abort();
      return;
   }

}
//---------------------------------------------------------------------------------------
DDHUndoableKnife::~DDHUndoableKnife()
{

   if( mUndoCalled )
   {
      // Remove the new objects
      for( List *obs = mPostKnife; obs != NULL; obs = obs->next )
      {
         t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
         object_free( ob->object );
      }
   }
   else
   {
      // Destroy data - the objects are no longer present
      object_free( mDeletedKnife );
      // Remove the old objects
      for( List *obs = mPreKnife; obs != NULL; obs = obs->next )
      {
         t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
         object_free( ob->object );
      }
   }

   list_free( &mPostKnife );
   list_free( &mPreKnife );
   list_free( &mPostSelectedObjects );

}
//---------------------------------------------------------------------------------------
Vertex *DDHUndoableKnife::intersectEdgeByTriangle( Vertex *newvertex, ACEdge *edge,
                                                   Surface *knifeSurf )
{

   newvertex = NULL;
   Vertex *kv0 = (( SVertex * )knifeSurf->vertlist->data )->v;

   // Segment direction
   Vertex dir;
   dir.x = edge->v2->x - edge->v1->x;
   dir.y = edge->v2->y - edge->v1->y;
   dir.z = edge->v2->z - edge->v1->z;
   // Direction to triangle
   Vertex w0;
   w0.x = edge->v1->x - kv0->x;
   w0.y = edge->v1->y - kv0->y;
   w0.z = edge->v1->z - kv0->z;

   float a = -DDH_DOTPRODUCT( knifeSurf->normal, w0 );
   float b = DDH_DOTPRODUCT( knifeSurf->normal, dir );
   if( fabs( b ) < DDH_EPSILON )
   {
      // Segment is either on plane or does not intersect plane.
      // (N == 0 means the segment is on the plane --- otherwise, no intersection occurred.
      return NULL;
   }

   // Get the intersection
   float sI = a / b;
   //if( sI < ( 0.0f + DDH_SLICE_EPSILON ) || sI > ( 1.0f - DDH_SLICE_EPSILON ))
   //if( sI < ( 0.0f - DDH_SLICE_EPSILON ) || sI > ( 1.0f + DDH_SLICE_EPSILON ))
   if( sI <= 0.0f || sI >= 1.0f )
   {
      // No intersection (< 0.0 == ray goes away from tri, > 1.0 == segment ends before plane)
      return NULL;
   }

   // Get intersect point
   Vertex I;
   Vertex Ipt;
   Ipt.x = sI * dir.x;
   Ipt.y = sI * dir.y;
   Ipt.z = sI * dir.z;
   I.x = edge->v1->x + Ipt.x;
   I.y = edge->v1->y + Ipt.y;
   I.z = edge->v1->z + Ipt.z;

   // Get triangle edge vectors
   // TODO: Precalculate this section per-triangle for speed
   Vertex *kv1 = (( SVertex * )knifeSurf->vertlist->next->data )->v;
   Vertex *kv2 = (( SVertex * )knifeSurf->vertlist->next->next->data )->v;
   Vertex u, v;
   u.x = kv1->x - kv0->x;
   u.y = kv1->y - kv0->y;
   u.z = kv1->z - kv0->z;
   v.x = kv2->x - kv0->x;
   v.y = kv2->y - kv0->y;
   v.z = kv2->z - kv0->z;

   // Check for intersect with triangle
   Vertex w;
   float uu, uv, vv, wu, wv, D;
   uu = DDH_DOTPRODUCT( u, u );
   uv = DDH_DOTPRODUCT( u, v );
   vv = DDH_DOTPRODUCT( v, v );
   w.x = I.x - kv0->x;
   w.y = I.y - kv0->y;
   w.z = I.z - kv0->z;
   wu = DDH_DOTPRODUCT( w, u );
   wv = DDH_DOTPRODUCT( w, v );
   D = uv * uv - uu * vv;

   // Test parametric coords
   float s, t;
   s = ( uv * wv - vv * wu ) / D;
   if( s < 0.0 || s > 1.0 )
   {
      // No intersect
      return NULL;
   }
   t = ( uv * wu - uu * wv ) / D;
   if( t < 0.0 || ( s + t ) > 1.0 )
   {
      // No intersect
      return NULL;
   }

   /*TODO:
   // Determine whether the vertex is close enough to an existing vertex to be a match
   if( fabs( I.x - edge->v1->x ) < DDH_SLICE_EPSILON &&
       fabs( I.y - edge->v1->y ) < DDH_SLICE_EPSILON &&
       fabs( I.z - edge->v1->z ) < DDH_SLICE_EPSILON )
   {
      return edge->v1;
   }
   else if( fabs( I.x - edge->v2->x ) < DDH_SLICE_EPSILON &&
            fabs( I.y - edge->v2->y ) < DDH_SLICE_EPSILON &&
            fabs( I.z - edge->v2->z ) < DDH_SLICE_EPSILON )
   {
      return edge->v2;
   }
   */

   // Add the vertex between the edge vertices
   Boolean wasnew;
   newvertex = ac_edge_insert_vertex( edge, &wasnew );
   // Get the % distance between the new vertex and its final position
   double newdist = sqrtf( pow( edge->v1->x - I.x, 2 ) +
                           pow( edge->v1->y - I.y, 2 ) +
                           pow( edge->v1->z - I.z, 2 ));
   double origdist = sqrtf( pow( edge->v1->x - edge->v2->x, 2 ) +
                            pow( edge->v1->y - edge->v2->y, 2 ) +
                            pow( edge->v1->z - edge->v2->z, 2 ));
   double pctdist = fabs( newdist ) / fabs( origdist );
   // Adjust the vertex coordinates
   for( List *sl = edge->surfacelist; sl != NULL; sl = sl->next )
   {
      Surface *s = ( Surface * )( sl->data );
      SVertex *sthis = surface_get_svertex( s, newvertex );
      SVertex *sfirst = surface_get_svertex( s, edge->v1 );
      SVertex *ssecond = surface_get_svertex( s, edge->v2 );
      sthis->tx = sfirst->tx - (( sfirst->tx - ssecond->tx ) * pctdist );
      sthis->ty = sfirst->ty - (( sfirst->ty - ssecond->ty ) * pctdist );
   }
   newvertex->x = I.x;
   newvertex->y = I.y;
   newvertex->z = I.z;

   return newvertex;

}
//---------------------------------------------------------------------------------------
void DDHUndoableKnife::sliceSurfaces( List *newvertlist )
{

   // Get our pool of potential surfaces
   List *surfpool = NULL;
   for( List *p = newvertlist; p != NULL; p = p->next )
   {
      list_append_list_unique( &surfpool, surfaces_of_vertex(( Vertex * )p->data ));
   }

   // Iterate through surfaces, finding valid surfaces to slice
   bool foundInvalidSlice = false;
   for( List *p = surfpool; p != NULL; p = p->next )
   {
      Surface *s = ( Surface * )p->data;
      // Identify the vertices in this surface that will form the slice
      Vertex *v1 = NULL;
      Vertex *v2 = NULL;
      int vord1 = -1;
      int vord2 = -1;
      int ctr = 0;
      for( List *p2 = s->vertlist; p2 != NULL; p2 = p2->next )
      {
         SVertex *sv = ( SVertex * )p2->data;
         for( List *p3 = newvertlist; p3 != NULL; p3 = p3->next )
         {
            Vertex *v = ( Vertex * )p3->data;
            if( sv->v == v )
            {
               if( vord1 == -1 )
               {
                  vord1 = ctr;
                  v1 = ( Vertex * )p2->data;
               }
               else if( vord2 == -1 )
               {
                  vord2 = ctr;
                  v2 = ( Vertex * )p2->data;
               }
               else
               {
                  // Found 3 verts in the same surface
                  foundInvalidSlice = true;
               }
               break;
            }
         }
         ++ctr;
      }
      // Only one vertex on surface selected - ignore slice
      if( v2 == NULL || v1 == NULL )
      {
      }
      // Ensure the vertices are not sequential
      else if( vord2 - vord1 == 1 || ( vord1 == 0 && vord2 == s->numvert - 1 ))
      {
         foundInvalidSlice = true;
      }
      else
      {
         // This is a valid slice surface - slice it
         // Add new surf #1
         ACObject *obj = object_of_surface( s );
         Surface *newsurf = new_surface();
         int lctr = 0;
         for( List *svp = s->vertlist; svp != NULL; svp = svp->next )
         {
            if( lctr >= vord1 && lctr <= vord2 )
            {
               SVertex *sv = ( SVertex * )svp->data;
               surface_add_vertex( newsurf, sv->v, sv->tx, sv->ty );
            }
            ++lctr;
         }
         surface_copy_properties( s, newsurf );
         object_add_surface( obj, newsurf );
         // Add new surf #2
         newsurf = new_surface();
         lctr = 0;
         for( List *svp = s->vertlist; svp != NULL; svp = svp->next )
         {
            if( lctr >= vord2 )
            {
               SVertex *sv = ( SVertex * )svp->data;
               surface_add_vertex( newsurf, sv->v, sv->tx, sv->ty );
            }
            ++lctr;
         }
         lctr = 0;
         for( List *svp = s->vertlist; svp != NULL; svp = svp->next )
         {
            if( lctr <= vord1 )
            {
               SVertex *sv = ( SVertex * )svp->data;
               surface_add_vertex( newsurf, sv->v, sv->tx, sv->ty );
            }
            ++lctr;
         }
         surface_copy_properties( s, newsurf );
         object_add_surface( obj, newsurf );
         // Remove old surface
         object_delete_surface( obj, s );
      }
   }

   // Had to remove this warning because now taht we're doing multiple splits
   // per edge, we may end up doing a split on a line where two vertices are
   // adjacent.
   /*
   if( foundInvalidSlice )
   {
      message_dialog( "knife : not all surfaces were sliced.  Surfaces can only be sliced where exactly two non-adjacent vertices are added." );
   }
   */

   list_free( &surfpool );

}
//---------------------------------------------------------------------------------------
bool DDHUndoableKnife::vertInsideKnife( Vertex *v, List *knifetris )
{
   // Return false if vertex is on/outside any of the knife triangles
   for( List *pk = knifetris; pk != NULL; pk = pk->next )
   {
      Surface *s = ( Surface * )pk->data;
      Vertex *v1 = (( SVertex * )s->vertlist->data )->v;
      Vertex *v2 = (( SVertex * )s->vertlist->next->data )->v;
      Vertex *v3 = (( SVertex * )s->vertlist->next->next->data )->v;
      // Get distance to plane
      // TODO: precalculate plane_d
      float plane_d = -( s->normal.x * v1->x + s->normal.y * v1->y + s->normal.z * v1->z );
      float distance = ( s->normal.x * v->x + s->normal.y * v->y + s->normal.z * v->z ) +
                         plane_d;
      if( distance > DDH_EPSILON )
      {
         return false;
      }
   }
   return true;
}
//---------------------------------------------------------------------------------------
void DDHUndoableKnife::cutAwaySurfaces( List *knifetris )
{

   // List of selected surfaces
   List *selsurf = ac_selection_get_whole_surfaces_all();
   if( selsurf == NULL )
   {
      message_dialog( "knife : no whole surfaces selected - no surfaces to cut away" );
      return;
   }

   // List of surfaces to cut away
   List *cutsurfs = NULL;

   // For each selected surfaces, determine where vertices exist in "front" of
   // any of the knife surfaces
   for( List *psurf = selsurf; psurf != NULL; psurf = psurf->next )
   {
      Surface *s = ( Surface * )psurf->data;
      bool isInside = true;
      for( List *pvert = s->vertlist; pvert != NULL; pvert = pvert->next )
      {
         SVertex *v = ( SVertex * )pvert->data;
         if( !vertInsideKnife( v->v, knifetris ))
         {
            isInside = false;
            break;
         }
      }
      if( isInside )
      {
         list_add_item_head( &cutsurfs, s );
      }
   }

   if( cutsurfs == NULL )
   {
      message_dialog( "knife : no surfaces were found to cut away; all surfaces are either on the same side of the knife, or the knife has sides facing in too many directions.  Try a knife that has a CSG/solid shape." );
      return;
   }

   // Clear the old select list
   list_free( &mPostSelectedObjects );
   mPostSelectedObjects = NULL;
   clear_selection();
   // Cut away surfaces per-object
   int old_mode = ac_get_select_mode();
   clear_selection();
   set_new_select_mode( SELECT_SURFACE );
   display_message( "Cutting away surfaces" );
   List *newObjects = NULL;
   for( List *pk = mPostKnife; pk != NULL; pk = pk->next )
   {

      t_mKnifeTargetObject *t = ( t_mKnifeTargetObject * )pk->data;

      // Identify the cutaway surfaces that belong to this object
      List *osurfs = ac_object_get_surfacelist( t->object );
      List *osurfscut = NULL;
      int ictr = ddhlist_merge( &osurfscut, cutsurfs, osurfs );

      // All surfaces to be cut away from this object have been culled into osurfscut,
      if( ictr == 0 )
      {
         // No surfaces were outside the knife
         continue;
      }
      else if( ictr == list_count( osurfs ))
      {
         // All surfaces were outside the knife
         list_add_item_head( &mPostSelectedObjects, t->object );
         list_free( &osurfscut );
         continue;
      }

      // Copy the cutaway surfaces

      // Only some surfaces in this object were outside the knife.
      // Separate these into their own object.
      List *newsurfs = NULL;
      List *vlist = NULL;
      clear_selection();
      ac_selection_select_surfacelist( osurfscut );
      List *copyList = clone_selected( TRUE );
      List *removeSurfaces = NULL;
      for( List *p = osurfscut; p != NULL; p = p->next )
      {
         Surface *s = ( Surface * )p->data;
         for( List *i = s->vertlist; i != NULL; i = i->next )
         {
            // note SVERTEX gets the actual vertex!
            list_add_item_unique( &vlist, SVERTEX( i->data ));
         }
         list_add_item_head( &removeSurfaces, s );
      }
      //object_remove_surfacelist( t->object, removeSurfaces );
      for( List *spt = removeSurfaces; spt != NULL; spt = spt->next )
      {
         object_remove_surface( t->object, ( Surface * )( spt->data ));
      }
      ac_surfacelist_free( &removeSurfaces );
      list_free( &removeSurfaces );

      for( List *p = vlist; p != NULL; p = p->next )
      {
         List *s = surfaces_of_vertex(( Vertex * )p->data );
         if( s != NULL )
         {
            list_free( &s );
         }
         else
         {
            object_delete_vertex( t->object, ( Vertex * )p->data );
         }
      }
      list_free( &vlist );

      object_add_children( t->parent, copyList );
      // Add the newly created object to the undo lists
      list_append_list_unique( &mPostSelectedObjects, copyList );
      t_mKnifeTargetObject *kto = new t_mKnifeTargetObject;
      kto->parent = t->parent;
      kto->object = ( ACObject * )( copyList->data );
      list_add_item_head( &newObjects, kto );
      list_free( &copyList );

      object_calc_normals( t->object );
      list_free( &osurfscut );

   }
   list_append_list_unique( &mPostKnife, newObjects );
   set_new_select_mode( old_mode );
   ac_selection_select_objectlist( mPostSelectedObjects );
   list_free( &cutsurfs );

}
//---------------------------------------------------------------------------------------
void DDHUndoableKnife::execute()
{

   // Get the objects we're potentially affecting
   List *currlist = ac_selection_get_objects();
   // Clone the objects
   for( List *obs = currlist; obs != NULL; obs = obs->next )
   {
      // Get the existing object
      ACObject *ob1 = ( ACObject * )obs->data;
      // Ignore objects with kids
      if( ac_object_get_childrenlist( ob1 ) != NULL )
      {
         continue;
      }
      // Get the object's parent
      ACObject *obpar = object_parent( ob1 );
      // Create a "pre" entry
      t_mKnifeTargetObject *oOld = new t_mKnifeTargetObject;
      oOld->parent = obpar;
      oOld->object = ob1;
      list_add_item_head( &mPreKnife, oOld );
      // Clone the object
      ACObject *ob2 = object_clone( ob1 );
      // Create a "post" entry
      t_mKnifeTargetObject *oNew = new t_mKnifeTargetObject;
      oNew->object = ob2;
      oNew->parent = obpar;
      list_add_item_head( &mPostKnife, oNew );
   }
   // Remove the old from the document
   for( List *obs = mPreKnife; obs != NULL; obs = obs->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
      object_remove_child_nocleanup( ob->parent, ob->object );
   }
   // Add the new objects to the scene, select them
   for( List *pk = mPostKnife; pk != NULL; pk = pk->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )pk->data;
      object_add_child( ob->parent, ob->object );
      list_add_item_head( &mPostSelectedObjects, ob->object );
   }
   ac_selection_select_objectlist( mPostSelectedObjects );

   int edgesSliced = 0;

   // List of knife surfaces
   //List *surflist = ac_object_get_triangle_surfaces( mKnifeObject );
   List *surflist = ac_object_get_surfacelist( mKnifeObject );

   // Keep slicing until there are no more edges to slice.  This way we can slice edges
   // that are intersected more than once
   bool sliceOccurred;
   // Number of times we've sliced surfaces
   int iterations = 0;
   // Maximum number of tries before we're done
   int maxIterations = list_count( surflist );
   // TODO: For performance, try identifying all edges up front and throw
   //       out any edges that could not possibly be sliced by the selected
   //       knife (outside bounding box range)
   for( List *thissurflist = surflist; thissurflist != NULL; thissurflist = thissurflist->next )
   {
      Surface *thissurf = ( Surface * )( thissurflist->data );
      display_message( "Slicing objects, pass %d/%d", iterations + 1, maxIterations );
      List *tempvertlist = NULL;
      sliceOccurred = false;
      // Get the surfaces to cut
      List *seledges = ac_selection_get_edges( false );
      //List *seledges = NULL;
      //DDHSelectionGetEdges( &seledges );
      if( seledges == NULL )
      {
         // Nothing to do
         return;
      }
      // Loop through the edges of our target
      for( List *p = seledges; p != NULL; p = p->next )
      {
         ACEdge *edge = ( ACEdge * )p->data;
         Vertex *newvertex = NULL;
         // Loop through the triangulated surfaces of our knife
         List *tslist = surface_get_triangulations( thissurf );
         for( List *pk = tslist; pk != NULL; pk = pk->next )
         {
            Surface *s = ( Surface * )pk->data;
            newvertex = intersectEdgeByTriangle( newvertex, edge, s );
            if( newvertex != NULL )
            {
               sliceOccurred = true;
               ++edgesSliced;
               // Add the vertex to our temp list (for slicing surfaces later)
               list_add_item_unique( &tempvertlist, newvertex );
               // We only allow one intersect per edge
               break;
            }
         }
         list_free( &tslist );
      }
      ++iterations;
      ac_edgelist_free( &seledges );
      // Slice the surfaces
      if( mSliceSurfacesInd )
      {
         sliceSurfaces( tempvertlist );
      }
      list_free( &tempvertlist );
   }

   // TODO:
   // If we're going to implement cutting surfaces inside a polygon, this is the place
   // to do it --- we've already sliced up all surfaces.

   display_message( "Calculating normals" );
   selected_calc_normals();

   // Cut away the surfaces
   if( mCutAwayInd )
   {
      List *tsurflist = ac_object_get_triangle_surfaces( mKnifeObject );
      cutAwaySurfaces( tsurflist );
      list_free( &tsurflist );
   }

   // Remove the knife
   if( mDeleteKnifeInd )
   {
      mDeletedKnife = mKnifeObject;
      mDeletedKnifeOwner = object_parent( mDeletedKnife );
      object_remove_child_nocleanup( mDeletedKnifeOwner, mDeletedKnife );
   }

   redraw_all();

   display_status();
   display_message( "Edges sliced: %d", edgesSliced );

}
//---------------------------------------------------------------------------------------
void DDHUndoableKnife::undo()
{

   mUndoCalled = true;

   // Recreate the knife
   if( mDeleteKnifeInd )
   {
      object_add_child( mDeletedKnifeOwner, mDeletedKnife );
   }

   // Remove the new objects
   for( List *obs = mPostKnife; obs != NULL; obs = obs->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
      object_remove_child_nocleanup( ob->parent, ob->object );
   }
   // Add the old objects
   for( List *obs = mPreKnife; obs != NULL; obs = obs->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
      object_add_child( ob->parent, ob->object );
   }
   // Select the old
   List *templist = NULL;
   for( List *pk = mPreKnife; pk != NULL; pk = pk->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )pk->data;
      list_add_item_head( &templist, ob->object );
   }
   ac_selection_select_objectlist( templist );
   list_free( &templist );

   redraw_all();

}
//---------------------------------------------------------------------------------------
void DDHUndoableKnife::redo()
{

   mUndoCalled = false;

   // Remove the knife
   if( mDeleteKnifeInd )
   {
      object_remove_child_nocleanup( mDeletedKnifeOwner, mDeletedKnife );
   }

   // Remove the old objects
   for( List *obs = mPreKnife; obs != NULL; obs = obs->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
      object_remove_child_nocleanup( ob->parent, ob->object );
   }
   // Add the new objects
   for( List *obs = mPostKnife; obs != NULL; obs = obs->next )
   {
      t_mKnifeTargetObject *ob = ( t_mKnifeTargetObject * )obs->data;
      object_add_child( ob->parent, ob->object );
   }
   // Select the new
   ac_selection_select_objectlist( mPostSelectedObjects );

   redraw_all();

}
//---------------------------------------------------------------------------------------

